@@ -18,7 +18,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  * 
  */
-#define RELEASE "0.10.1"
+#define RELEASE "0.11"
 
 /* Includes {{{ */
 #include <stdio.h>
@@ -68,7 +68,16 @@ static GtkWidget *infoLabelBox;
 static GtkWidget *mouseEventBox;
 static GdkPixbuf *currentImage = NULL;
 static GdkPixbuf *scaledImage = NULL;
+
+#ifndef NO_ANIMATIONS
+static GdkPixbufAnimation *currentAnimation = NULL;
+static GSList *animationImageBuffer = NULL;
+static GdkPixbufAnimationIter *animationIterator = NULL;
+static GdkPixbufAnimation *memoryArgAnimation = NULL;
+#else
 static GdkPixbuf *memoryArgImage = NULL;
+#endif
+
 static gchar emptyCursor[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
 /* Inline images: The chess board like one for transparent images and
@@ -126,6 +135,10 @@ static gshort slideshowInterval = 3;
 static gchar slideshowEnabled = 0;
 static guint slideshowID = 0;
 static gchar aliases[128];
+#ifndef NO_ANIMATIONS
+static guint animationTimeoutID = 0;
+static gboolean currentImageIsAnimated = FALSE;
+#endif
 #ifndef NO_INOTIFY
 static int inotifyFd = 0;
 static int inotifyWd = -1;
@@ -160,6 +173,7 @@ gboolean reloadImage();
 void autoScaleFactor();
 void resizeAndPosWindow();
 void displayImage();
+inline void scale();
 /* }}} */
 /* Error, debug and info message stuff {{{ */
 /* Debugging {{{ */
@@ -432,7 +446,11 @@ gboolean loadFiles(gchar **iterator) { /*{{{*/
 	while(*iterator != 0) {
 		if(strcmp(*iterator, "-") == 0) {
 			/* Load image from stdin {{{ */
+			#ifndef NO_ANIMATIONS
+			if(memoryArgAnimation != NULL) {
+			#else
 			if(memoryArgImage != NULL) {
+			#endif
 				g_printerr("You can't specify more than one image to "
 					"be read from stdin.\n");
 				iterator++;
@@ -453,15 +471,25 @@ gboolean loadFiles(gchar **iterator) { /*{{{*/
 					g_error_free(loadError);
 					g_object_unref(memoryImageLoader);
 					loadError = NULL;
-					memoryArgImage = (GdkPixbuf*)1; /* Ignore further
-						attempts to load an image from stdin */
+					#ifndef NO_ANIMATIONS
+					memoryArgAnimation = (GdkPixbufAnimation*)1; /* Ignore further attempts
+						to load an image from stdin */
+					#else
+					memoryArgImage = (GdkPixbuf*)1; /* Ignore further attempts
+						to load an image from stdin */
+					#endif
 					break;
 				}
 			}
 			if(gdk_pixbuf_loader_close(memoryImageLoader, &loadError) == TRUE) {
+				#ifndef NO_ANIMATIONS
+				memoryArgAnimation = gdk_pixbuf_loader_get_animation(memoryImageLoader);
+				// Don't unref as we can't copy animations
+				#else
 				memoryArgImage = gdk_pixbuf_copy(
 					gdk_pixbuf_loader_get_pixbuf(memoryImageLoader));
 				g_object_unref(memoryImageLoader);
+				#endif
 				loadFilesAddFile("-");
 			}
 			else {
@@ -469,8 +497,13 @@ gboolean loadFiles(gchar **iterator) { /*{{{*/
 					loadError->message);
 				g_error_free(loadError);
 				g_object_unref(memoryImageLoader);
+				#ifndef NO_ANIMATIONS
+				memoryArgAnimation = (GdkPixbufAnimation*)1; /* Ignore further attempts
+					to load an image from stdin */
+				#else
 				memoryArgImage = (GdkPixbuf*)1; /* Ignore further attempts
 					to load an image from stdin */
+				#endif
 			}
 			g_free(buf);
 			iterator++;
@@ -622,6 +655,14 @@ void runProgram(gchar *command) { /*{{{*/
 		/* }}} */
 	}
 	else if(command[0] == '|') {
+		#ifndef NO_ANIMATIONS
+		if(currentImageIsAnimated) {
+			setInfoText("Command piping is not supported for animations");
+			return;
+		}
+		#endif
+
+
 		/* Pipe data {{{ */
 		childArgv[0] = "/bin/sh";
 		childArgv[1] = "-c";
@@ -702,6 +743,7 @@ void runProgram(gchar *command) { /*{{{*/
 			g_free(buf);
 			exit(0);
 		} /* }}} */
+		wait(NULL);
 	}
 } /*}}}*/
 #endif
@@ -770,37 +812,168 @@ void sortFiles(int (*compareFunction)(const void*, const void*)) {
 #endif
 /* }}} */
 /*}}}*/
+#ifndef NO_ANIMATIONS
+/* Animated images support {{{ */
+gboolean animationIterateCb(gpointer data) {
+	int delayTime;
+	gboolean mustRedraw = FALSE;
+
+	if(!currentAnimation) {
+		animationTimeoutID = 0;
+		return FALSE;
+	}
+
+	// Iterate
+	mustRedraw = gdk_pixbuf_animation_iter_advance(animationIterator, NULL);
+
+	if(mustRedraw) {
+		// Load new image
+		currentImage = gdk_pixbuf_animation_iter_get_pixbuf(animationIterator);
+		scaledAt = -1;
+		scale();
+		displayImage();
+	}
+
+	delayTime = gdk_pixbuf_animation_iter_get_delay_time(animationIterator);
+	if(delayTime != -1) {
+		animationTimeoutID = g_timeout_add(delayTime, animationIterateCb, NULL);
+	}
+	return FALSE;
+}
+/* }}} */
+#endif
 /* Load images and modify them {{{ */
 gboolean loadImage() { /*{{{*/
 	/**
 	 * Load an image and display it
 	 */
 	GdkPixbuf *tmpImage;
+	GdkPixbuf *tmpImage2;
 	GdkPixbuf *chessBoardBuf;
+	#ifndef NO_ANIMATIONS
+	GdkPixbufAnimation *tmpAnimation;
+	#endif
 	GError *error = NULL;
 	gint i, n, o, p;
 
 	DEBUG2("loadImage", currentFile->fileName);
 	if(strcmp(currentFile->fileName, "-") == 0) {
 		/* Load memory image */
+		#ifndef NO_ANIMATIONS
+		if(memoryArgAnimation == NULL) {
+			return FALSE;
+		}
+		tmpAnimation = g_object_ref(memoryArgAnimation);
+		if(gdk_pixbuf_animation_is_static_image(tmpAnimation)) {
+			tmpImage2 = gdk_pixbuf_apply_embedded_orientation(gdk_pixbuf_animation_get_static_image(tmpAnimation));
+			tmpImage = gdk_pixbuf_copy(tmpImage2);
+			g_object_unref(tmpImage2);
+			if(!tmpImage) {
+				g_printerr("Failed to load %s\n", currentFile->fileName);
+				g_object_unref(tmpAnimation);
+				return FALSE;
+			}
+			g_object_unref(tmpAnimation);
+			currentImageIsAnimated = FALSE;
+		}
+		else {
+			currentImageIsAnimated = TRUE;
+			DEBUG1("Current image is animated");
+		}
+		#else
 		if(memoryArgImage == NULL) {
 			return FALSE;
 		}
 		tmpImage = g_object_ref(memoryArgImage);
+		#endif
 	}
 	else {
 		/* Load from file */
+		#ifndef NO_ANIMATIONS
+		tmpAnimation = gdk_pixbuf_animation_new_from_file(currentFile->fileName, &error);
+		if(!tmpAnimation) {
+			g_printerr("Failed to load %s: %s\n", currentFile->fileName, error->message);
+			g_error_free(error);
+			return FALSE;
+		}
+		if(gdk_pixbuf_animation_is_static_image(tmpAnimation)) {
+			tmpImage2 = gdk_pixbuf_apply_embedded_orientation(gdk_pixbuf_animation_get_static_image(tmpAnimation));
+			tmpImage = gdk_pixbuf_copy(tmpImage2);
+			g_object_unref(tmpImage2);
+			if(!tmpImage) {
+				g_printerr("Failed to load %s\n", currentFile->fileName);
+				g_object_unref(tmpAnimation);
+				return FALSE;
+			}
+			g_object_unref(tmpAnimation);
+			currentImageIsAnimated = FALSE;
+		}
+		else {
+			currentImageIsAnimated = TRUE;
+			DEBUG1("Current image is animated");
+		}
+		#else
 		tmpImage = gdk_pixbuf_new_from_file(currentFile->fileName, &error);
 		if(!tmpImage) {
 			g_printerr("Failed to load %s: %s\n", currentFile->fileName, error->message);
 			g_error_free(error);
 			return FALSE;
 		}
+		#endif
 	}
+	// Free old, now unused stuff
+	#ifndef NO_ANIMATIONS
+	if(currentAnimation != NULL) {
+		g_object_unref(currentAnimation);
+		currentImage = NULL;
+		currentAnimation = NULL;
+	}
+	if(animationImageBuffer != NULL) {
+		g_slist_free(animationImageBuffer);
+		animationImageBuffer = NULL;
+	}
+	if(animationIterator != NULL) {
+		g_object_unref(animationIterator);
+		animationIterator = NULL;
+	}
+	if(animationTimeoutID != 0) {
+		g_source_remove(animationTimeoutID);
+		animationTimeoutID = 0;
+	}
+	#endif
 	if(currentImage != NULL) {
-		DEBUG1("Free old image");
-		g_object_unref(currentImage);
+		if(G_IS_OBJECT(currentImage)) {
+			g_object_unref(currentImage);
+		}
+		currentImage = NULL;
 	}
+	#ifndef NO_ANIMATIONS
+	if(currentImageIsAnimated) {
+		currentAnimation = tmpAnimation;
+
+		// Setup iterator
+		animationIterator = gdk_pixbuf_animation_get_iter(tmpAnimation, NULL);
+
+		// Get image to-be-displayed
+		tmpImage = gdk_pixbuf_animation_iter_get_pixbuf(animationIterator);
+
+		// Setup image cache
+		animationImageBuffer = g_slist_alloc();
+
+		// Setup animation callback function
+		o = gdk_pixbuf_animation_iter_get_delay_time(animationIterator);
+		if(o != -1) {
+			animationTimeoutID = g_timeout_add(o, animationIterateCb, NULL);
+		}
+
+		// We don't support chessboards / autorotation for animated images (yet)
+		currentImage = tmpImage;
+		zoom = 1;
+		moveX = moveY = 0;
+		scaledAt = -1;
+		return TRUE;
+	}
+	#endif
 	if(optionHideChessboardLevel == 0 && gdk_pixbuf_get_has_alpha(tmpImage)) {
 		/* Draw chessboard for transparent images */
 		DEBUG1("Creating chessboard");
@@ -826,17 +999,6 @@ gboolean loadImage() { /*{{{*/
 		currentImage = tmpImage;
 	}
 
-	/* Rotate image */
-	#if GDK_PIXBUF_MAJOR >= 2 && GDK_PIXBUF_MINOR >= 12
-	if(gdk_pixbuf_get_option(currentImage, "orientation") != NULL) {
-		tmpImage = gdk_pixbuf_apply_embedded_orientation(currentImage);
-		g_object_unref(currentImage);
-		currentImage = tmpImage;
-	}
-	#else
-		#warning "Automatic image rotation requires GDK Pixbuf >= 2.12. Feature disabled!"
-	#endif
-
 	#ifndef NO_INOTIFY
 	/* Update inotify watch to manage automatic reloading {{{ */ 
 	if(optionUseInotify == TRUE) {
@@ -869,10 +1031,17 @@ inline void scale() { /*{{{*/
 		imgy = gdk_pixbuf_get_height(currentImage);
 		scaledAt = zoom;
 		if(scaledImage != NULL) {
-			DEBUG1("Free");
-			g_object_unref(scaledImage);
+			if(G_IS_OBJECT(scaledImage)) {
+				DEBUG1("Free");
+				g_object_unref(scaledImage);
+			}
 			scaledImage = NULL;
 		}
+		if(scaledAt == 1) {
+			// No need to calculate. Just copy.
+			scaledImage = g_object_ref(currentImage);
+			return;
+		}
 		if(imgx > 10 && imgy > 10 && (imgx * zoom < 10 || imgy * zoom < 10)) {
 			/* Don't zoom below 10x10 pixels (except for images that actually are that small) */
 			scaledImage = gdk_pixbuf_scale_simple(currentImage, 10, 10,
@@ -926,6 +1095,15 @@ void forceAutoScaleFactor(enum autoScaleSetting upDown) { /*{{{*/
 		}
 	}
 
+	#ifndef NO_ANIMATIONS
+	// Don't autoscale animations up
+	// This is absolutely required. Animations lag heavily at
+	// high zoom levels.
+	if(currentImageIsAnimated && zoom > 1) {
+		zoom = 1;
+	}
+	#endif
+
 	/* Now do actually scale */
 	scale();
 } /*}}}*/
@@ -1682,6 +1860,12 @@ gboolean keyboardCb(GtkWidget *widget, GdkEventKey *event, gpointer data) { /*{{
 		/* }}} */
 		/* BIND: l: Rotate left {{{ */
 		case GDK_l:
+			#ifndef NO_ANIMATIONS
+			if(currentImageIsAnimated) {
+				setInfoText("Rotation not supported for animations");
+				break;
+			}
+			#endif
 			setInfoText("Rotated left");
 			rotate(TRUE);
 			autoScaleFactor();
@@ -1691,6 +1875,12 @@ gboolean keyboardCb(GtkWidget *widget, GdkEventKey *event, gpointer data) { /*{{
 			/* }}} */
 		/* BIND: k: Rotate right {{{ */
 		case GDK_k:
+			#ifndef NO_ANIMATIONS
+			if(currentImageIsAnimated) {
+				setInfoText("Rotation not supported for animations");
+				break;
+			}
+			#endif
 			setInfoText("Rotated right");
 			rotate(FALSE);
 			autoScaleFactor();
@@ -1700,6 +1890,12 @@ gboolean keyboardCb(GtkWidget *widget, GdkEventKey *event, gpointer data) { /*{{
 			/* }}} */
 		/* BIND: h: Horizontal flip {{{ */
 		case GDK_h:
+			#ifndef NO_ANIMATIONS
+			if(currentImageIsAnimated) {
+				setInfoText("Flipping not supported for animations");
+				break;
+			}
+			#endif
 			setInfoText("Flipped horizontally");
 			flip(TRUE);
 			displayImage();
@@ -1707,6 +1903,12 @@ gboolean keyboardCb(GtkWidget *widget, GdkEventKey *event, gpointer data) { /*{{
 			/* }}} */
 		/* BIND: v: Vertical flip {{{ */
 		case GDK_v:
+			#ifndef NO_ANIMATIONS
+			if(currentImageIsAnimated) {
+				setInfoText("Flipping not supported for animations");
+				break;
+			}
+			#endif
 			setInfoText("Flipped vertically");
 			flip(FALSE);
 			displayImage();
@@ -1972,6 +2174,7 @@ gboolean screenChangedCb(GtkWidget *widget, GdkScreen *previous_screen, gpointer
 	autoScaleFactor();
 	resizeAndPosWindow();
 	displayImage();
+	setInfoText(NULL);
 	return FALSE;
 }/*}}}*/
 gboolean receivedInitialFullscreen = FALSE;
@@ -2005,6 +2208,7 @@ gboolean windowStateCb(GtkWidget *widget, GdkEventWindowState *event, gpointer d
 		autoScaleFactor();
 		resizeAndPosWindow();
 		displayImage();
+		setInfoText(NULL);
 	}
 	return FALSE;
 }/*}}}*/
@@ -2053,15 +2257,18 @@ int main(int argc, char *argv[]) {
 		sprintf(fileName, "%s/.pqivrc", constBuf);
 
 		if(g_file_get_contents(fileName, &buf, NULL, NULL)) {
-			if(!g_shell_parse_argv(buf, &optionFileArgc, &optionFileArgv, NULL)) {
-				die("Failed to parse the options file ~/.pqivrc");
-			}
-			for(i=0; i < optionFileArgc; i++) {
-				options[optionCount] = (char*)malloc(strlen(optionFileArgv[i]));
-				g_stpcpy(options[optionCount], optionFileArgv[i]);
-				if(++optionCount > 250) {
-					die("Too many options; your configuration file "
-						"is restricted to 250 options");
+			g_strstrip(buf);
+			if(buf[0] != 0) {
+				if(!g_shell_parse_argv(buf, &optionFileArgc, &optionFileArgv, NULL)) {
+					die("Failed to parse the options file ~/.pqivrc");
+				}
+				for(i=0; i < optionFileArgc; i++) {
+					options[optionCount] = (char*)malloc(strlen(optionFileArgv[i]));
+					g_stpcpy(options[optionCount], optionFileArgv[i]);
+					if(++optionCount > 250) {
+						die("Too many options; your configuration file "
+							"is restricted to 250 options");
+					}
 				}
 			}
 			/* We can't free due to a bug in ubuntu's GTK version
@@ -2134,7 +2341,11 @@ int main(int argc, char *argv[]) {
 			/* OPTION: -r: Read additional filenames (not folders) from stdin */
 			case 'r':
 				optionReadStdin = TRUE;
+				#ifndef NO_ANIMATIONS
+				memoryArgAnimation = (GdkPixbufAnimation*)1; /* Don't allow - files */
+				#else
 				memoryArgImage = (GdkPixbuf*)1; /* Don't allow - files */
+				#endif
 				break;
 			/* OPTION: -c: Disable the background for transparent images */
 			#ifndef NO_COMPOSITING
